# pylint: disable=missing-docstring


import textwrap
import unittest
from typing import Optional, MutableMapping

import tests.common
import tests.infer_for_schema.common
from aas_core_codegen import infer_for_schema, intermediate


class Test_property_in_set(unittest.TestCase):
    def test_no_constraints(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"

            class Something:
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # fmt: off
        (
            _,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            textwrap.dedent(
                """\
                ConstraintsByProperty(
                  len_constraints_by_property={},
                  patterns_by_property={},
                  set_of_primitives_by_property={},
                  set_of_enumeration_literals_by_property={})"""
            ),
            text,
        )

    def test_property_in_set(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])

            @invariant(
                lambda self: self.some_property in Some_set,
                "Some property must be part of some set."
            )
            class Something:
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # fmt: off
        (
            _,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral A',
        'Reference to EnumerationLiteral B'])})""",
            text,
        )

    def test_property_in_set_in_conjunction(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])

            Another_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.B,
                    Some_enum.C
                ])


            @invariant(
                lambda self:
                self.some_property in Some_set
                and self.some_property in Another_set,
                "Some property must be part of some set and another set."
            )
            class Something:
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # fmt: off
        (
            _,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral B'])})""",
            text,
        )

    def test_property_in_set_in_implication(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])

            @invariant(
                lambda self:
                not (self.some_property is not None)
                or self.some_property in Some_set,
                "Some property must be part of some set."
            )
            class Something:
                some_property: Optional[Some_enum]

                def __init__(self, some_property: Optional[Some_enum] = None) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # fmt: off
        (
            _,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral A',
        'Reference to EnumerationLiteral B'])})""",
            text,
        )

    def test_property_in_set_in_implication_and_conjunction(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])

            Another_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.B,
                    Some_enum.C
                ])

            @invariant(
                lambda self:
                not (self.some_property is not None)
                or (
                    self.some_property in Some_set
                    and self.some_property in Another_set
                ),
                "Some property must be part of some set and another set."
            )
            class Something:
                some_property: Optional[Some_enum]

                def __init__(self, some_property: Optional[Some_enum] = None) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # fmt: off
        (
            _,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral B'])})""",
            text,
        )


class Test_stacking(unittest.TestCase):
    def test_only_inherited_and_no_constraints_of_its_own(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])


            @invariant(
                lambda self:
                self.some_property in Some_set,
                "Some property must be part of some set."
            )
            class Parent(DBC):
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            class Something(Parent):
                def __init__(self, some_property: Some_enum) -> None:
                    Parent.__init__(self, some_property)

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # noinspection PyUnusedLocal
        constraints_by_class: Optional[
            MutableMapping[
                intermediate.ClassUnion, infer_for_schema.ConstraintsByProperty
            ]
        ] = None  # Necessary for mypy

        # fmt: off
        (
            symbol_table,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_class, error = infer_for_schema.merge_constraints_with_ancestors(
            symbol_table=symbol_table, constraints_by_class=constraints_by_class
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert constraints_by_class is not None

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral A',
        'Reference to EnumerationLiteral B'])})""",
            text,
        )

    def test_merge_with_parent(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B
                ])

            Another_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.B,
                    Some_enum.C
                ])

            @invariant(
                lambda self:
                self.some_property in Some_set,
                "Some property must be part of some set."
            )
            class Parent(DBC):
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            @invariant(
                lambda self:
                self.some_property in Another_set,
                "Some property must be part of another set."
            )
            class Something(Parent):
                def __init__(self, some_property: Some_enum) -> None:
                    Parent.__init__(self, some_property)

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # noinspection PyUnusedLocal
        constraints_by_class: Optional[
            MutableMapping[
                intermediate.ClassUnion, infer_for_schema.ConstraintsByProperty
            ]
        ] = None  # Necessary for mypy

        # fmt: off
        (
            symbol_table,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_class, error = infer_for_schema.merge_constraints_with_ancestors(
            symbol_table=symbol_table, constraints_by_class=constraints_by_class
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert constraints_by_class is not None

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral B'])})""",
            text,
        )

    def test_merge_with_parent_and_grand_parent(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                A = "A"
                B = "B"
                C = "C"
                D = "D"
                E = "E"

            Some_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.A,
                    Some_enum.B,
                    Some_enum.C,
                ])

            Another_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.B,
                    Some_enum.C,
                    Some_enum.D
                ])

            Yet_another_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.C,
                    Some_enum.D,
                    Some_enum.E
                ])

            @invariant(
                lambda self:
                self.some_property in Some_set,
                "Some property must be part of some set."
            )
            class Grand_parent(DBC):
                some_property: Some_enum

                def __init__(self, some_property: Some_enum) -> None:
                    self.some_property = some_property

            @invariant(
                lambda self:
                self.some_property in Another_set,
                "Some property must be part of another set."
            )
            class Parent(Grand_parent):
                def __init__(self, some_property: Some_enum) -> None:
                    Grand_parent.__init__(self, some_property)

            @invariant(
                lambda self:
                self.some_property in Yet_another_set,
                "Some property must be part of yet another set."
            )
            class Something(Parent):
                def __init__(self, some_property: Some_enum) -> None:
                    Parent.__init__(self, some_property)


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        # noinspection PyUnusedLocal
        constraints_by_class: Optional[
            MutableMapping[
                intermediate.ClassUnion, infer_for_schema.ConstraintsByProperty
            ]
        ] = None  # Necessary for mypy

        # fmt: off
        (
            symbol_table,
            something_cls,
            constraints_by_class,
        ) = (
            tests.infer_for_schema.common
            .parse_to_symbol_table_and_something_cls_and_constraints_by_class(
                source=source
            )
        )
        # fmt: on

        constraints_by_class, error = infer_for_schema.merge_constraints_with_ancestors(
            symbol_table=symbol_table, constraints_by_class=constraints_by_class
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert constraints_by_class is not None

        constraints_by_props = constraints_by_class[something_cls]

        text = infer_for_schema.dump(constraints_by_props)
        self.assertEqual(
            """\
ConstraintsByProperty(
  len_constraints_by_property={},
  patterns_by_property={},
  set_of_primitives_by_property={},
  set_of_enumeration_literals_by_property={
    'some_property': SetOfEnumerationLiteralsConstraint(
      enumeration='Reference to Enumeration Some_enum',
      literals=[
        'Reference to EnumerationLiteral C'])})""",
            text,
        )


if __name__ == "__main__":
    unittest.main()



"""
Provide a meta-model of something.

:constraint AAS-001:
    Some constraint
"""


class Something:
    """
    Represent something.

    :constraint AAS-001:
        Duplicate constraint

    """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""
Update everything in this project to the latest aas-core-meta.

Git is expected to be installed.
"""

import argparse
import os
import pathlib
import subprocess
import sys
import tempfile
import time
from typing import Optional

import dev_scripts.download_latest_aas_core_meta_v3


def _make_sure_no_changed_files(
    repo_dir: pathlib.Path, expected_our_branch: str
) -> Optional[int]:
    """
    Make sure that no files are modified in our repository.

    Return exit code if something is unexpected.
    """
    diff_name_status = subprocess.check_output(
        ["git", "diff", "--name-status", expected_our_branch],
        cwd=str(repo_dir),
        encoding="utf-8",
    ).strip()

    if len(diff_name_status.splitlines()) > 0:
        print(
            f"The following files are modified "
            f"compared to branch {expected_our_branch!r} in {repo_dir}:\n"
            f"{diff_name_status}\n"
            f"\n"
            f"Please stash the changes first before you update to aas-core-meta.",
            file=sys.stderr,
        )
        return 1

    return None


def _rerecord_everything(repo_dir: pathlib.Path) -> Optional[int]:
    """
    Run all unit tests with re-record environment variable set.

    Return an error code, if any.
    """
    env = os.environ.copy()
    env["AAS_CORE_CODEGEN_RERECORD"] = "1"

    starting_points = [
        pth
        for pth in (repo_dir / "tests").glob("*")
        if (
            pth.is_dir()
            and not pth.name.startswith("__")
            and not pth.name.startswith(".")
        )
    ]

    start = time.perf_counter()

    for starting_point in starting_points:
        print(f"Starting to run tests in: {starting_point} ...")

        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "unittest",
                "discover",
                "--start-directory",
                str(starting_point),
            ],
            env=env,
            cwd=str(repo_dir),
        )

    duration = time.perf_counter() - start
    print(f"Re-recording took: {duration:.2f} seconds.")
    return None


def _create_branch_commit_and_push(
    repo_dir: pathlib.Path, aas_core_meta_revision: str
) -> None:
    """Create a feature branch, commit the changes and push it."""
    branch = f"Update-test-data-to-aas-core-meta-{aas_core_meta_revision}"
    print(f"Creating the branch {branch!r}...")
    subprocess.check_call(["git", "checkout", "-b", branch], cwd=repo_dir)

    print("Adding files...")
    subprocess.check_call(["git", "add", "."], cwd=repo_dir)

    # pylint: disable=line-too-long
    message = f"""\
Update test data to aas-core-meta {aas_core_meta_revision}

We update the development requirements to and re-record the test data
for [aas-core-meta {aas_core_meta_revision}].

[aas-core-meta {aas_core_meta_revision}]: https://github.com/aas-core-works/aas-core-meta/commit/{aas_core_meta_revision}"""
    # pylint: enable=line-too-long

    print("Committing...")
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_file = pathlib.Path(tmp_dir) / "commit-message.txt"
        tmp_file.write_text(message, encoding="utf-8")

        subprocess.check_call(["git", "commit", "--file", str(tmp_file)], cwd=repo_dir)

    print(f"Pushing to remote {branch}...")
    subprocess.check_call(["git", "push", "-u"], cwd=repo_dir)


def main() -> int:
    """Execute the main routine."""
    repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--expected_our_branch",
        help="Git branch expected in this repository",
        default="main",
    )
    args = parser.parse_args()

    expected_our_branch = str(args.expected_our_branch)

    our_branch = subprocess.check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=str(repo_dir),
        encoding="utf-8",
    ).strip()
    if our_branch != expected_our_branch:
        print(
            f"--expected_our_branch is {expected_our_branch}, "
            f"but got {our_branch} in: {repo_dir}",
            file=sys.stderr,
        )
        return 1

    exit_code = _make_sure_no_changed_files(
        repo_dir=repo_dir, expected_our_branch=expected_our_branch
    )
    if exit_code is not None:
        return exit_code

    # fmt: off
    aas_core_meta_revision = (
        dev_scripts.download_latest_aas_core_meta_v3.retrieve_sha_and_download()
    )[:8]
    # fmt: on

    print(f"The aas-core-meta revision is: {aas_core_meta_revision}")

    exit_code = _rerecord_everything(repo_dir=repo_dir)
    if exit_code is not None:
        return exit_code

    _create_branch_commit_and_push(
        repo_dir=repo_dir, aas_core_meta_revision=aas_core_meta_revision
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



# pylint: disable=missing-docstring

import unittest

import aas_core_codegen.golang.common as golang_common


class TestStringLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual('""', golang_common.string_literal(""))

    def test_no_quotes(self) -> None:
        self.assertEqual('"x"', golang_common.string_literal("x"))

    def test_double_quotes_but_no_single_quotes(self) -> None:
        self.assertEqual('"a\\"b"', golang_common.string_literal('a"b'))

    def test_single_quotes_but_no_double_quotes(self) -> None:
        self.assertEqual('"a\'b"', golang_common.string_literal("a'b"))

    def test_tab(self) -> None:
        self.assertEqual('"a\\tb"', golang_common.string_literal("a\tb"))


class TestBytesLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertTupleEqual(
            ("[...]byte{}", False), golang_common.bytes_literal(bytearray(b""))
        )

    def test_one(self) -> None:
        self.assertTupleEqual(
            ("[...]byte{0x00}", False),
            golang_common.bytes_literal(bytearray(b"\x00")),
        )

    def test_eight(self) -> None:
        self.assertTupleEqual(
            ("[...]byte{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07}", False),
            golang_common.bytes_literal(bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07")),
        )

    def test_nine(self) -> None:
        code, multiline = golang_common.bytes_literal(
            bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08")
        )

        self.assertEqual(
            """\
[...]byte {
\t0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
\t0x08
}""",
            code,
        )
        self.assertTrue(multiline)

    def test_sixteen(self) -> None:
        code, multiline = golang_common.bytes_literal(
            bytearray(
                b"\x00\x01\x02\x03\x04\x05\x06\x07" b"\x08\x09\x10\x11\x12\x13\x14\x15"
            )
        )

        self.assertEqual(
            """\
[...]byte {
\t0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
\t0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15
}""",
            code,
        )

        self.assertTrue(multiline)


if __name__ == "__main__":
    unittest.main()



"""
Provide a meta-model of something.

:constraint AAS-001:
    Some constraint
"""


class Something:
    """
    Represent something.

    See :constraintref:`AAS-NON-EXISTING`
    """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""De-serialize the model, validate it and serialize it back."""
import argparse
import difflib
import json
import pathlib
import sys
from xml.dom import minidom

if sys.version_info >= (3, 9):
    from typing import Literal, Optional, Any
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

# The module aas_core is expected to be generated by the integration test.
from sdk import types, jsonization, xmlization, verification


def normalize_xml(xml_text: str) -> str:
    """Normalize XML for comparison."""
    normalized = minidom.parseString(xml_text).toprettyxml(indent=" ")
    # Remove blank lines
    return "\n".join([line for line in normalized.splitlines() if line.strip()])


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_path", help="Path to the model in JSON or XML", required=True
    )
    args = parser.parse_args()

    model_path = pathlib.Path(args.model_path)

    if not model_path.exists():
        print(f"--model_path does not exist: {model_path}", file=sys.stderr)
        sys.exit(1)

    text = model_path.read_text(encoding="utf-8")
    stripped_text = text.strip()

    if len(stripped_text) == 0:
        print(f"Model {model_path} is empty.", file=sys.stderr)
        sys.exit(1)

    # NOTE (mristin):
    # We assume by convention that the main object in the integration test is defined as
    # the class ``Root``.
    root: types.Root

    first_character = stripped_text[0]

    jsonable = None  # type: Optional[Any]

    source_format: Literal["json", "xml"]

    if first_character == "{":
        try:
            jsonable = json.loads(text)
        except json.JSONDecodeError as exception:
            print(
                f"Failed to parse JSON from {model_path}: {exception}", file=sys.stderr
            )
            sys.exit(1)

        try:
            root = jsonization.root_from_jsonable(jsonable)
        except jsonization.DeserializationException as exception:
            print(f"Failed to parse {model_path}: {exception}", file=sys.stderr)
            sys.exit(1)

        source_format = "json"

    elif first_character == "<":
        try:
            root = xmlization.root_from_str(text)
        except xmlization.DeserializationException as exception:
            print(f"Failed to parse {model_path}: {exception}", file=sys.stderr)
            sys.exit(1)

        source_format = "xml"

    else:
        print(
            f"Unrecognized first non-whitespace character in {model_path}, "
            f"so we do not know how to parse the file: {first_character!r}",
            file=sys.stderr,
        )
        sys.exit(1)

    for error in verification.verify(root):
        print(
            f"The model from {model_path} failed to pass verification "
            f"at {error.path}: {error.cause}",
            file=sys.stderr,
        )
        sys.exit(1)

    if source_format == "json":
        another_jsonable = jsonization.to_jsonable(root)

        assert jsonable is not None

        original_pretty = json.dumps(jsonable, indent=2, ensure_ascii=False)
        another_pretty = json.dumps(another_jsonable, indent=2, ensure_ascii=False)

        if original_pretty != another_pretty:
            diff = "\n".join(
                difflib.unified_diff(
                    original_pretty.splitlines(),
                    another_pretty.splitlines(),
                    fromfile="original",
                    tofile="roundtrip",
                    lineterm="",
                )
            )
            print(
                f"JSON round-trip mismatch for {model_path}:\n{diff}",
                file=sys.stderr,
            )
            sys.exit(1)

    elif source_format == "xml":
        assert isinstance(root, types.Root)

        another_text = xmlization.to_str(root)

        if normalize_xml(text) != normalize_xml(another_text):
            diff = "\n".join(
                difflib.unified_diff(
                    normalize_xml(text).splitlines(),
                    normalize_xml(another_text).splitlines(),
                    fromfile="original",
                    tofile="roundtrip",
                    lineterm="",
                )
            )
            print(
                f"XML round-trip mismatch for {model_path}:\n{diff}",
                file=sys.stderr,
            )
            sys.exit(1)

    else:
        assert_never(source_format)


if __name__ == "__main__":
    main()



Some_set: Set[str] = constant_set(values=["some literal", "another literal"])


@invariant(
    lambda self: self.some_property in Some_set,
    "Some property must belong to Some set.",
)
class Something:
    some_property: str

    def __init__(self, some_property: str) -> None:
        self.some_property = some_property


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""Run all the tests with the environment variable ``AAS_CORE_CODEGEN_RERECORD`` set."""

import argparse
import os
import pathlib
import subprocess
import sys
import shlex


def main() -> int:
    """Execute the main routine."""
    available_tests = [
        "tests.cpp.test_main.Test_against_recorded",
        "tests.cpp.test_verification.Test_against_recorded",
        "tests.csharp.test_main.Test_against_recorded",
        "tests.csharp.test_verification.Test_against_recorded",
        "tests.csharp.test_structure.Test_generation_against_recorded",
        "tests.intermediate.test_translate.Test_against_recorded",
        "tests.java.test_main.Test_against_recorded",
        "tests.our_jsonschema.test_main.Test_against_recorded",
        "tests.rdf_shacl.test_main.Test_against_recorded",
        "tests.parse.test_parse.Test_against_recorded",
        "tests.parse.test_retree.Test_against_recorded",
        "tests.python.test_main.Test_against_recorded",
        "tests.smoke.test_main.Test_against_recorded",
        "tests.typescript.test_main.Test_against_recorded",
        "tests.xsd.test_main.Test_against_recorded",
    ]

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--select",
        help=(
            "If set, only the selected tests are executed. "
            "This is practical if some of the tests failed and you want to "
            "fix them in isolation. "
            "The tests are given as a space-separated list of: "
            + " ".join(available_tests)
        ),
        metavar="",
        nargs="+",
        choices=available_tests,
    )
    args = parser.parse_args()

    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    dev_root = repo_root / "dev"

    env = os.environ.copy()
    env["AAS_CORE_CODEGEN_RERECORD"] = "1"

    if args.select is not None:
        tests_to_run = list(args.select)
    else:
        tests_to_run = available_tests

    for qualified_test_name in tests_to_run:
        cmd = [sys.executable, "-m", "unittest", "-v", qualified_test_name]

        print(f"Executing {qualified_test_name}...")
        exit_code = subprocess.call(cmd, cwd=str(dev_root), env=env)
        if exit_code != 0:
            cmd_str = " ".join(shlex.quote(part) for part in cmd)
            print(f"Failed to execute the test: {cmd_str}")
            return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



@invariant(lambda self: len(self) > 0, "Non-empty")
class Something(str):
    pass


@invariant(lambda self: len(self) <= 9, "At most 9 characters")
class Child(Something):
    pass


@invariant(lambda self: len(self) <= 5, "At most 5 characters")
class Grand_child(Child):
    pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Item:
    some_str: str

    def __init__(self, some_str: str) -> None:
        self.some_str = some_str


class Something:
    items: List[Item]
    optional_items: Optional[List[Item]]

    def __init__(
        self, items: List[Item], optional_items: Optional[List[Item]] = None
    ) -> None:
        self.items = items
        self.optional_items = optional_items


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# We test with this sample that the method signature is correctly translated.
@abstract
class SomethingAbstract:
    @require(lambda x: x > 0)
    def some_func(self, x: int) -> bool:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Modelling_kind(Enum):
    Template = "Template"
    Instance = "Instance"


class Something:
    modelling_kind: Modelling_kind
    optional_modelling_kind: Optional[Modelling_kind]

    def __init__(
        self,
        modelling_kind: Modelling_kind,
        optional_modelling_kind: Optional[Modelling_kind] = None
    ) -> None:
        self.modelling_kind = modelling_kind
        self.optional_modelling_kind = optional_modelling_kind


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# The invariants are contradicting.
@invariant(lambda self: len(self.some_prop) == 10)
@invariant(lambda self: len(self.some_prop) == 11)
class Something:
    some_prop: str

    def __init__(self, some_prop: str) -> None:
        self.some_prop = some_prop


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import unittest

from aas_core_codegen.cpp import common as cpp_common


class TestBreakTypeInLines(unittest.TestCase):
    def test_empty(self) -> None:
        result = cpp_common.break_type_in_lines("")
        self.assertEqual(result, "")

    def test_no_breaks(self) -> None:
        result = cpp_common.break_type_in_lines("const int")
        self.assertEqual("const int", result)

    def test_breaks(self) -> None:
        result = cpp_common.break_type_in_lines(
            "std::optional<std::vector<std::shared_ptr<IEmbeddedDataSpecification> > >&"
        )
        self.assertEqual(
            """\
std::optional<
  std::vector<
    std::shared_ptr<
      IEmbeddedDataSpecification > > >&""",
            result,
        )


if __name__ == "__main__":
    unittest.main()



import contextlib
import pathlib
import tempfile
import unittest
import io

from aas_core_codegen.main import Parameters, Target, execute

import tests.common
from tests.rdf_shacl.test_main import REPO_DIR


class Test_jsonld_context(unittest.TestCase):
    def test_jsonld_context_generation(self) -> None:
        for model_pth in tests.common.REAL_META_MODEL_PATHS:
            case_dir = (
                REPO_DIR / "dev" / "test_data" / "jsonld_context" / model_pth.stem
            )
            expected_output_dir = case_dir / "output"

            with contextlib.ExitStack() as exit_stack:
                if tests.common.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    # pylint: disable=consider-using-with
                    tmp_dir = tempfile.TemporaryDirectory()
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = Parameters(
                    model_path=model_pth,
                    target=Target.JSONLD_CONTEXT,
                    snippets_dir=case_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()
                return_code = execute(params=params, stdout=stdout, stderr=stderr)
                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                if not tests.common.RERECORD:
                    generated_jsonld_context_path = output_dir / "context.jsonld"
                    generated_jsonld_context = generated_jsonld_context_path.read_text(
                        encoding="utf-8"
                    )

                    expected_jsonld_context_path = (
                        expected_output_dir / "context.jsonld"
                    )
                    expected_jsonld_context = expected_jsonld_context_path.read_text(
                        encoding="utf-8"
                    )

                    self.assertEqual(
                        generated_jsonld_context,
                        expected_jsonld_context,
                        f"The generated and the expected context differ. "
                        f"The expected JSON-LD context lives "
                        f"at: {expected_jsonld_context_path}, while the generated one "
                        f"at: {generated_jsonld_context_path}",
                    )


if __name__ == "__main__":
    unittest.main()



class Something:
    @require(lambda self, x: self.y < x, description="some contract")
    def do_something(self, x: int) -> int:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import os
import pathlib
import unittest
from typing import List, Tuple, Optional

from aas_core_codegen.parse import retree as parse_retree
from aas_core_codegen.intermediate import revm as intermediate_revm

import tests.common


class Test_against_recorded(unittest.TestCase):
    def test_cases(self) -> None:
        repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent
        test_cases_dir = repo_root / "dev/test_data/intermediate_revm"

        assert test_cases_dir.exists(), f"{test_cases_dir=}"
        assert test_cases_dir.is_dir(), f"{test_cases_dir=}"

        # The expected cases should have no errors.
        expected_pths = sorted((test_cases_dir / "expected").glob("**/pattern.regex"))
        regex_pths_expected_exception = [
            (pth, False) for pth in expected_pths
        ]  # type: List[Tuple[pathlib.Path, bool]]

        unexpected_pths = sorted(
            (test_cases_dir / "unexpected").glob("**/pattern.regex")
        )
        regex_pths_expected_exception.extend((pth, True) for pth in unexpected_pths)

        for regex_pth, expected_exception in regex_pths_expected_exception:
            case_dir = regex_pth.parent

            try:
                regex_text = regex_pth.read_text(encoding="utf-8")
            except Exception as exception:
                raise AssertionError(
                    f"Unexpected exception when reading "
                    f"from {regex_pth.relative_to(repo_root)}"
                ) from exception

            try:
                regex, error = parse_retree.parse([regex_text])
            except Exception as exception:
                raise AssertionError(
                    f"Unexpected exception in parsing "
                    f"the regex pattern {regex_text!r} "
                    f"from {regex_pth.relative_to(repo_root)}"
                ) from exception

            if error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(error.cursor)
                raise AssertionError(
                    f"Unexpected error in parsing "
                    f"the regex pattern "
                    f"from {regex_pth.relative_to(repo_root)}:\n"
                    f"{error.message}:\n"
                    f"{regex_line}\n"
                    f"{pointer_line}"
                )

            assert regex is not None

            program = None  # type: Optional[intermediate_revm.NodeOrLeaf]
            caught = None  # type: Optional[Exception]

            try:
                program = intermediate_revm.translate(regex=regex)
            except Exception as exception:
                caught = exception

            if not expected_exception and caught is not None:
                raise AssertionError(
                    f"Expected no errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got:\n"
                    f"{caught}"
                ) from caught

            elif expected_exception and caught is None:
                raise AssertionError(
                    f"Expected exception in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got none."
                )

            else:
                pass

            expected_program_pth = case_dir / "program.txt"
            expected_exception_pth = case_dir / "exception.txt"

            if expected_exception:
                assert program is None
                assert caught is not None

                caught_str = str(caught)

                if tests.common.RERECORD:
                    expected_program_pth.unlink(missing_ok=True)
                    expected_exception_pth.write_text(caught_str, encoding="utf-8")
                else:
                    assert not expected_program_pth.exists(), (
                        f"Unexpected program stored in the test case "
                        f"where we expect an exception: {expected_program_pth}"
                    )
                    expected_caught_str = expected_exception_pth.read_text(
                        encoding="utf-8"
                    )
                    self.assertEqual(expected_caught_str, caught_str, f"{case_dir=}")

            else:
                assert program is not None
                assert caught is None

                program_str = intermediate_revm.dump(program)

                if tests.common.RERECORD:
                    expected_program_pth.write_text(program_str, encoding="utf-8")
                    expected_exception_pth.unlink(missing_ok=True)
                else:
                    assert not expected_exception_pth.exists(), (
                        f"Unexpected exception stored in the test case "
                        f"where we expect a program: {expected_exception_pth}"
                    )

                    expected_program_str = expected_program_pth.read_text(
                        encoding="utf-8"
                    )
                    self.assertEqual(expected_program_str, program_str, f"{case_dir=}")


if __name__ == "__main__":
    unittest.main()



"""Provide common functionality across different tests."""
import os
import pathlib
from typing import List, Tuple, Optional, Union, Sequence, Final

import asttokens
from icontract import ensure, require

from aas_core_codegen import parse, intermediate
from aas_core_codegen.common import Error


# pylint: disable=missing-function-docstring


def most_underlying_messages(error_or_errors: Union[Error, Sequence[Error]]) -> str:
    """Find the "leaf" errors and render them as a new-line separated list."""
    if isinstance(error_or_errors, Error):
        errors = [error_or_errors]  # type: Sequence[Error]
    else:
        errors = error_or_errors

    most_underlying_errors = []  # type: List[Error]

    for error in errors:
        if error.underlying is None or len(error.underlying) == 0:
            most_underlying_errors.append(error)
            continue

        stack = error.underlying  # type: List[Error]

        while len(stack) > 0:
            top_error = stack.pop()

            if top_error.underlying is not None:
                stack.extend(top_error.underlying)

            if top_error.underlying is None or len(top_error.underlying) == 0:
                most_underlying_errors.append(top_error)

    return "\n".join(
        most_underlying_error.message
        for most_underlying_error in most_underlying_errors
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def parse_atok(
    atok: asttokens.ASTTokens,
) -> Tuple[Optional[parse.SymbolTable], Optional[Error]]:
    """Parse the ``atok``, an abstract syntax tree of a meta-model."""
    import_errors = parse.check_expected_imports(atok=atok)
    if len(import_errors) > 0:
        import_errors_str = "\n".join(
            f"* {import_error}" for import_error in import_errors
        )

        raise AssertionError(
            f"Unexpected imports in the source code:\n{import_errors_str}"
        )

    symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    return symbol_table, error


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def parse_source(source: str) -> Tuple[Optional[parse.SymbolTable], Optional[Error]]:
    """Parse the given source text into a symbol table."""
    atok, parse_exception = parse.source_to_atok(source=source)
    if parse_exception:
        raise parse_exception  # pylint: disable=raising-bad-type

    assert atok is not None

    return parse_atok(atok=atok)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def translate_source_to_intermediate(
    source: str,
) -> Tuple[Optional[intermediate.SymbolTable], Optional[Error]]:
    atok, parse_exception = parse.source_to_atok(source=source)
    if parse_exception:
        raise parse_exception  # pylint: disable=raising-bad-type

    assert atok is not None

    parsed_symbol_table, error = parse_atok(atok=atok)
    assert error is None, f"{most_underlying_messages(error)}"
    assert parsed_symbol_table is not None

    return intermediate.translate(parsed_symbol_table=parsed_symbol_table, atok=atok)


def must_translate_source_to_intermediate(
    source: str,
) -> intermediate.SymbolTable:
    atok, parse_exception = parse.source_to_atok(source=source)
    if parse_exception:
        raise parse_exception  # pylint: disable=raising-bad-type

    assert atok is not None

    parsed_symbol_table, error = parse_atok(atok=atok)
    assert error is None, f"{most_underlying_messages(error)}"
    assert parsed_symbol_table is not None

    symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table, atok=atok
    )
    assert (
        error is None
    ), f"Unexpected error when parsing the source: {most_underlying_messages(error)}"
    assert symbol_table is not None
    return symbol_table


#: If set, this environment variable indicates that the golden files should be
#: re-recorded instead of checked against.
RERECORD = os.environ.get("AAS_CORE_CODEGEN_RERECORD", "").lower() in (
    "1",
    "true",
    "on",
)


class TestCaseWithDirectoryAndMetaModel:
    """Represent a test case with files in a test directory and a meta-model."""

    #: Path to the expected files and other test data
    case_dir: Final[pathlib.Path]

    #: Path to the meta-model file corresponding to the test case
    model_path: Final[pathlib.Path]

    def __init__(self, case_dir: pathlib.Path, model_path: pathlib.Path) -> None:
        """Initialize with the given values."""
        self.case_dir = case_dir
        self.model_path = model_path


@require(lambda base_case_dir: base_case_dir.exists() and base_case_dir.is_dir())
def test_cases_from_base_case_dir(
    base_case_dir: pathlib.Path,
) -> List[TestCaseWithDirectoryAndMetaModel]:
    """
    Find meta-model files beneath the ``base_case_dir``.

    We resolve the meta-model source by checking for each test case directory if
    the file ``meta_model.py`` exists.

    :param base_case_dir: parent directory where test cases reside
    :return: list of found test cases
    """
    test_cases = []  # type: List[TestCaseWithDirectoryAndMetaModel]

    for case_dir in sorted(pth for pth in base_case_dir.iterdir() if pth.is_dir()):
        model_pth = case_dir / "meta_model.py"  # type: Optional[pathlib.Path]
        assert model_pth is not None

        if model_pth.exists():
            test_cases.append(
                TestCaseWithDirectoryAndMetaModel(
                    case_dir=case_dir, model_path=model_pth
                )
            )
            continue

    return test_cases


@require(lambda base_case_dir: base_case_dir.exists() and base_case_dir.is_dir())
def test_cases_from_real_world_models(
    base_case_dir: pathlib.Path, real_meta_model_paths: Sequence[pathlib.Path]
) -> List[TestCaseWithDirectoryAndMetaModel]:
    """
    Find the test cases for the given real-world models.

    The meta-model path will point to the real-world model, which is shared across many
    different tests, while the case directory will be situated beneath
    the ``base_case_dir`` corresponding to the file name of the real-world model.
    """
    return [
        TestCaseWithDirectoryAndMetaModel(
            case_dir=base_case_dir / model_pth.stem, model_path=model_pth
        )
        for model_pth in real_meta_model_paths
    ]


def _repo_root() -> pathlib.Path:
    return pathlib.Path(os.path.realpath(__file__)).parent.parent.parent


REAL_META_MODEL_PATHS: Final[Sequence[pathlib.Path]] = [
    _repo_root() / "dev/test_data/real_meta_models/aas_core_meta.v3.py"
]



# We explicitly test here that the constructor is not included in the signatures.
@abstract
class SomethingAbstract:
    x: int

    def __init__(self, x: int) -> None:
        self.x = x


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



from typing import List


class My_Class:
    string_prop: str
    int_prop: int
    bool_prop: bool

    def __init__(self, string_prop: str, int_prop: int, bool_prop: bool) -> None:
        self.string_prop = string_prop
        self.int_prop = int_prop
        self.bool_prop = bool_prop


class Root:
    foo: My_Class
    foos: List[My_Class]
    strings: List[str]
    integers: List[int]
    booleans: List[bool]

    def __init__(
        self,
        foo: My_Class,
        foos: List[My_Class],
        strings: List[str],
        integers: List[int],
        booleans: List[bool],
    ) -> None:
        self.foo = foo
        self.foos = foos
        self.strings = strings
        self.integers = integers
        self.booleans = booleans


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""Check that metadata in ``pyproject.toml`` and ``__init__.py`` are consistent."""

import importlib.metadata
import sys
from typing import List, TypeVar, Protocol, Iterator, Union, Any, Optional, cast

import aas_core_codegen

# NOTE (mristin):
# We found the official PackageMetadata protocol to be unreliable, so we wrote our
# own, closer to the actual implementation.

_T = TypeVar("_T")


class PackageMetadata(Protocol):  # pylint: disable=missing-docstring
    def get_all(self, name: str, failobj: _T = ...) -> Union[List[Any], _T]:
        """
        Return all values associated with a possibly multi-valued key.
        """

    def __getitem__(self, key: str) -> Optional[str]:
        ...  # pragma: no cover

    def __iter__(self) -> Iterator[str]:
        ...  # pragma: no cover

    def __contains__(self, item: str) -> bool:
        ...  # pragma: no cover

    def __len__(self) -> int:
        ...  # pragma: no cover


def _extract_authors(metadata: PackageMetadata) -> List[str]:
    """Extract author names listed in the metadata."""
    author_email = metadata["Author-Email"]

    authors = []
    if author_email is not None:
        # We parse "Name <email>, Name <email>" format.
        for entry in author_email.split(","):
            entry = entry.strip()
            if "<" in entry:
                name = entry.split("<")[0].strip()
                if name:
                    authors.append(name)
            else:
                authors.append(entry)

    # We also check Author field as a fallback.
    try:
        author = metadata["Author"]
    except KeyError:
        author = None

    if author is not None and len(authors) == 0:
        authors = [author.strip()]

    return authors


def _extract_status_from_classifiers(
    metadata: PackageMetadata,
) -> str:
    """Extract development status from classifiers listed in the metadata."""
    classifiers = metadata.get_all("Classifier") or []

    for classifier in classifiers:
        assert isinstance(classifier, str)

        if classifier.startswith("Development Status"):
            parts = classifier.split("::")
            if len(parts) >= 2:
                status = parts[1].strip()
                # We extract just the status part (e.g.,
                # "5 - Production/Stable" -> "Production/Stable")
                if " - " in status:
                    return status.split(" - ", 1)[1]
                return status
    return ""


def _extract_license_from_classifiers(
    metadata: PackageMetadata,
) -> str:
    """Extract license from classifiers listed in the metadata."""
    classifiers = metadata.get_all("Classifier") or []

    for classifier in classifiers:
        assert isinstance(classifier, str)

        if classifier.startswith("License"):
            return classifier
    return ""


def main() -> None:
    """Execute the main routine."""
    errors = []

    metadata = cast(PackageMetadata, importlib.metadata.metadata("aas-core-codegen"))

    # Check version
    init_version = aas_core_codegen.__version__
    pyproject_toml_version = metadata["version"]

    if init_version != pyproject_toml_version:
        errors.append(
            f"Version mismatch: __init__.py has {init_version!r}, "
            f"pyproject.toml has {pyproject_toml_version!r}"
        )

    # Check authors
    init_authors = [author.strip() for author in aas_core_codegen.__author__.split(",")]
    pyproject_authors = _extract_authors(metadata=metadata)

    if init_authors != pyproject_authors:
        errors.append(
            f"Authors mismatch: __init__.py has {init_authors!r}, "
            f"pyproject.toml has {pyproject_authors!r}"
        )

    # Check development status
    init_status = aas_core_codegen.__status__
    pyproject_status = _extract_status_from_classifiers(metadata=metadata)

    if init_status != pyproject_status:
        errors.append(
            f"Status mismatch: __init__.py has {init_status!r}, "
            f"pyproject.toml classifiers have {pyproject_status!r}"
        )

    # Check license
    init_license = aas_core_codegen.__license__
    pyproject_license = _extract_license_from_classifiers(metadata=metadata)

    if init_license != pyproject_license:
        errors.append(
            f"License mismatch: __init__.py has {init_license!r}, "
            f"pyproject.toml classifiers have {pyproject_license!r}"
        )

    if errors:
        print("Metadata consistency check failed:", file=sys.stderr)
        for error in errors:
            print(f"  {error}", file=sys.stderr)
        sys.exit(1)

    print("All meta-data fields are consistent between __init__.py and pyproject.toml")


if __name__ == "__main__":
    main()



"""
Copy the generated C++ files for AAS v3.0 to the corresponding C++ repository.

This script is practical as a part of the development pipeline:
1) Re-generate C++ files (through re-recording the test data),
2) Copy the generated files (with this script),
3) Run clang-format and clang-tidy from the C++ repository, and
4) Run the unit tests from the C++ repository.
"""

import argparse
import os
import pathlib
import shutil
import sys


def main() -> int:
    """Execute the main routine."""
    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--aas_core3_cpp_path",
        help="Path to the aas-core3.0-cpp repository",
        default=str(repo_root.parent / "aas-core3.0-cpp"),
    )
    args = parser.parse_args()

    aas_core3_cpp_path = pathlib.Path(args.aas_core3_cpp_path)

    if not aas_core3_cpp_path.exists():
        print(
            f"--aas_core3_cpp_path does not exist: {aas_core3_cpp_path}",
            file=sys.stderr,
        )
        return 1

    if not aas_core3_cpp_path.is_dir():
        print(
            f"--aas_core3_cpp_path is not a directory: {aas_core3_cpp_path}",
            file=sys.stderr,
        )
        return 1

    expected_output_dir = (
        repo_root
        / "dev"
        / "test_data"
        / "cpp"
        / "test_main"
        / "aas_core_meta.v3"
        / "expected_output"
    )

    assert aas_core3_cpp_path.exists()

    target_src_dir = aas_core3_cpp_path / "src"
    target_src_dir.mkdir(exist_ok=True)

    target_include_dir = aas_core3_cpp_path / "include" / "aas_core" / "aas_3_0"
    target_include_dir.mkdir(exist_ok=True, parents=True)

    paths = sorted(
        list(expected_output_dir.glob("*.cpp"))
        + list(expected_output_dir.glob("*.hpp"))
    )

    print(
        f"Copying *.cpp and *.hpp\n"
        f"  from {expected_output_dir}\n"
        f"  to {target_src_dir} and {target_include_dir}..."
    )
    for pth in paths:
        if pth.suffix == ".cpp":
            dst = target_src_dir / pth.name
        elif pth.suffix == ".hpp":
            dst = target_include_dir / pth.name
        else:
            raise AssertionError(f"Unhandled suffix: {pth.suffix} from: {pth}")

        # NOTE (mristin, 2024-01-08):
        # We compare the contents to avoid re-compilations by CMake.
        should_copy = True
        if dst.exists():
            src_text = pth.read_text(encoding="utf-8")
            dst_text = dst.read_text(encoding="utf-8")

            should_copy = src_text != dst_text

        if should_copy:
            shutil.copy(src=pth, dst=dst)

    print("Copied.")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# pylint: disable=missing-docstring

import unittest

from aas_core_codegen.common import Identifier
from aas_core_codegen.yielding import (
    flow as yielding_flow,
)
from aas_core_codegen.cpp import yielding as cpp_yielding


class TestExecuteBody(unittest.TestCase):
    def test_empty(self) -> None:
        code = cpp_yielding.generate_execute_body(
            flow=[], state_member=Identifier("state_")
        )
        self.assertEqual("// Intentionally empty.", code)

    def test_inspired_by_verificator(self) -> None:
        # NOTE (mristin, 2023-10-22):
        # We test inspired by a real-world example from C++ code generation so that
        # we can check whether the resulting code is readable.
        #
        # Hence, we test with two invariants and two constrained primitive properties
        # so that the example is complex enough.
        code = cpp_yielding.generate_execute_body(
            flow=[
                yielding_flow.IfFalse(
                    condition="CheckInvariant0()",
                    body=[
                        yielding_flow.command_from_text(
                            "error_ = ErrorForInvariant0();"
                        ),
                        yielding_flow.Yield(),
                    ],
                ),
                yielding_flow.IfFalse(
                    condition="CheckInvariant1()",
                    body=[
                        yielding_flow.command_from_text(
                            "error_ = ErrorForInvariant1();"
                        ),
                        yielding_flow.Yield(),
                    ],
                ),
                yielding_flow.IfTrue(
                    condition="instance_.prop0.HasValue()",
                    body=[
                        yielding_flow.command_from_text(
                            """\
verification = VerifyConstrainedPrimitive(instance_.prop0.Value());
iterator_ = verification.Begin();
iterator_end_ = verification.End();"""
                        ),
                        yielding_flow.For(
                            condition="iterator_ != iterator_end_",
                            iteration="++iterator_",
                            body=[
                                yielding_flow.command_from_text("error_ = *iterator_;"),
                                yielding_flow.Yield(),
                            ],
                        ),
                        yielding_flow.command_from_text(
                            """\
iterator_.reset();
iterator_end_.reset();"""
                        ),
                    ],
                ),
                yielding_flow.IfTrue(
                    condition="instance_.prop1.HasValue()",
                    body=[
                        yielding_flow.command_from_text(
                            """\
verification = VerifyConstrainedPrimitive(instance_.prop1.Value());
iterator_ = verification.Begin();
iterator_end_ = verification.End();"""
                        ),
                        yielding_flow.For(
                            condition="iterator_ != iterator_end_",
                            iteration="++iterator_",
                            body=[
                                yielding_flow.command_from_text("error_ = *iterator_;"),
                                yielding_flow.Yield(),
                            ],
                        ),
                        yielding_flow.command_from_text(
                            """\
iterator_.reset();
iterator_end_.reset();"""
                        ),
                    ],
                ),
                yielding_flow.command_from_text(
                    """\
error_.Reset();
Finalize();"""
                ),
            ],
            state_member=Identifier("state_"),
        )

        self.maxDiff = None
        self.assertEqual(
            """\
while (true) {
  switch (state_) {
    case 0: {
      if (CheckInvariant0()) {
        state_ = 1;
        continue;
      }

      error_ = ErrorForInvariant0();

      state_ = 1;
      return;
    }

    case 1: {
      if (CheckInvariant1()) {
        state_ = 2;
        continue;
      }

      error_ = ErrorForInvariant1();

      state_ = 2;
      return;
    }

    case 2: {
      if (!(instance_.prop0.HasValue())) {
        state_ = 6;
        continue;
      }

      verification = VerifyConstrainedPrimitive(instance_.prop0.Value());
      iterator_ = verification.Begin();
      iterator_end_ = verification.End();
    }

    case 3: {
      if (!(iterator_ != iterator_end_)) {
        state_ = 5;
        continue;
      }

      error_ = *iterator_;

      state_ = 4;
      return;
    }

    case 4: {
      ++iterator_

      state_ = 3;
      continue;
    }

    case 5: {
      iterator_.reset();
      iterator_end_.reset();
    }

    case 6: {
      if (!(instance_.prop1.HasValue())) {
        state_ = 10;
        continue;
      }

      verification = VerifyConstrainedPrimitive(instance_.prop1.Value());
      iterator_ = verification.Begin();
      iterator_end_ = verification.End();
    }

    case 7: {
      if (!(iterator_ != iterator_end_)) {
        state_ = 9;
        continue;
      }

      error_ = *iterator_;

      state_ = 8;
      return;
    }

    case 8: {
      ++iterator_

      state_ = 7;
      continue;
    }

    case 9: {
      iterator_.reset();
      iterator_end_.reset();
    }

    case 10: {
      error_.Reset();
      Finalize();

      // We invalidate the state since we reached the end of the routine.
      state_ = 11;
      return;
    }

    default:
      throw std::logic_error(
        common::Concat(
          "Invalid state_: ",
          std::to_string(state_)
        )
      );
  }
}""",
            code,
        )


if __name__ == "__main__":
    unittest.main()



class Something:
    some_bool: bool
    some_int: int
    some_float: float
    some_str: str
    some_byte_array: bytearray

    some_optional_bool: Optional[bool]
    some_optional_int: Optional[int]
    some_optional_float: Optional[float]
    some_optional_str: Optional[str]
    some_optional_byte_array: Optional[bytearray]

    def __init__(
        self,
        some_bool: bool,
        some_int: int,
        some_float: float,
        some_str: str,
        some_byte_array: bytearray,
        some_optional_bool: Optional[bool] = None,
        some_optional_int: Optional[int] = None,
        some_optional_float: Optional[float] = None,
        some_optional_str: Optional[str] = None,
        some_optional_byte_array: Optional[bytearray] = None,
    ) -> None:
        self.some_bool = some_bool
        self.some_int = some_int
        self.some_float = some_float
        self.some_str = some_str
        self.some_byte_array = some_byte_array
        self.some_optional_bool = some_optional_bool
        self.some_optional_int = some_optional_int
        self.some_optional_float = some_optional_float
        self.some_optional_str = some_optional_str
        self.some_optional_byte_array = some_optional_byte_array


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# ``with_model_type`` is missing in the serialization.
class Something:
    pass


class Concrete(Something):
    pass


class Bundle:
    something: Something

    def __init__(self, something: Something) -> None:
        self.something = something


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@invariant(lambda self: re.match(r"^hello$", self.x), "X must match hello.")
class Something:
    x: str

    def __init__(self, x: str) -> None:
        self.x = x


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Value:
    text: str

    def __init__(self, text: str) -> None:
        self.text = text


# ERROR: ``len(.)`` on a class is not defined.
@invariant(lambda self: len(self.value) > 1, "Value longer than 1")
class Something:
    value: Value

    def __init__(self, value: Value) -> None:
        self.value = value


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import ast
import os
import pathlib
import re
import textwrap
import unittest
from typing import Optional, Tuple, List

import asttokens
import docutils.nodes

from aas_core_codegen import parse
from aas_core_codegen.common import Error, Identifier

import tests.common


class Test_parsing_AST(unittest.TestCase):
    def test_valid_code(self) -> None:
        source = textwrap.dedent(
            """\
            class Something:
                pass
            """
        )

        atok, error = parse.source_to_atok(source=source)
        assert atok is not None
        assert error is None

    def test_invalid_code(self) -> None:
        source = textwrap.dedent(
            """\
            class Something: 12 this is wrong
            """
        )

        _, error = parse.source_to_atok(source=source)
        assert error is not None
        assert isinstance(error, SyntaxError)
        self.assertEqual(1, error.lineno)


class Test_checking_imports(unittest.TestCase):
    @staticmethod
    def replace_column_number_with_x(errors: List[str]) -> List[str]:
        """
        Replace the column number with ``"X"``.

        We need to remove the column number as it changes between Python versions
        (notably, it is more precise from Python 3.10 on).
        """
        return [re.sub(r"column [0-9]+", "column X", error) for error in errors]

    def test_import_reported(self) -> None:
        source = textwrap.dedent(
            """\
            import typing
            """
        )

        atok, error = parse.source_to_atok(source=source)
        assert error is None, f"{error=}"
        assert atok is not None

        errors = parse.check_expected_imports(atok=atok)
        self.assertListEqual(
            [
                "At line 1 and column X: "
                "Unexpected ``import ...``. "
                "Only ``from ... import...`` statements are expected."
            ],
            Test_checking_imports.replace_column_number_with_x(errors),
        )

    def test_from_import_as_reported(self) -> None:
        source = textwrap.dedent(
            """\
            from typing import List as Lst
            """
        )

        atok, error = parse.source_to_atok(source=source)
        assert error is None, f"{error=}"
        assert atok is not None

        # NOTE (mristin, 2022-01-22):
        # We need to remove the column number as it changes between Python versions
        # (notably, it is more precise from Python 3.10 on.)

        errors = parse.check_expected_imports(atok=atok)
        self.assertListEqual(
            [
                "At line 1 and column X: Unexpected ``from ... import ... as ...``. "
                "Only ``from ... import...`` statements are expected."
            ],
            Test_checking_imports.replace_column_number_with_x(errors),
        )

    def test_unexpected_name_from_module(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import List
            """
        )

        atok, error = parse.source_to_atok(source=source)
        assert atok is not None
        assert error is None

        errors = parse.check_expected_imports(atok=atok)
        self.assertListEqual(
            [
                "At line 1 and column X: "
                "Expected to import 'List' from the module typing, "
                "but it is imported from enum."
            ],
            Test_checking_imports.replace_column_number_with_x(errors),
        )

    def test_unexpected_import_from_a_module(self) -> None:
        source = textwrap.dedent(
            """\
            from something import Else
            """
        )

        atok, error = parse.source_to_atok(source=source)
        assert atok is not None
        assert error is None

        errors = parse.check_expected_imports(atok=atok)
        self.assertListEqual(
            ["At line 1 and column X: Unexpected import of a name 'Else'."],
            Test_checking_imports.replace_column_number_with_x(errors),
        )


class Test_parsing_docstring(unittest.TestCase):
    @staticmethod
    def parse_and_extract_docstring(source: str) -> docutils.nodes.document:
        """
        Parse the ``source`` and extract a description.

        The description is expected to belong to a single class, ``Some_class``.
        """
        symbol_table, error = tests.common.parse_source(source)
        assert error is None, f"{error}"
        assert symbol_table is not None

        cls = symbol_table.must_find_class(Identifier("Some_class"))
        assert cls.description is not None
        return cls.description.document

    def test_empty(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """"""

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        document = Test_parsing_docstring.parse_and_extract_docstring(source=source)

        self.assertEqual(0, len(document.children))

    def test_simple_single_line(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """This is some documentation."""

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        document = Test_parsing_docstring.parse_and_extract_docstring(source=source)

        self.assertEqual(1, len(document.children))
        self.assertIsInstance(document.children[0], docutils.nodes.paragraph)

    def test_that_multi_line_docstring_is_not_parsed_as_a_block_quote(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                This is some documentation.

                Another paragraph.
                """

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        document = Test_parsing_docstring.parse_and_extract_docstring(source=source)
        self.assertEqual(2, len(document.children))
        self.assertIsInstance(document.children[0], docutils.nodes.paragraph)
        self.assertIsInstance(document.children[1], docutils.nodes.paragraph)


class Test_unexpected_class_definitions(unittest.TestCase):
    @staticmethod
    def error_from_source(source: str) -> Optional[Error]:
        """Encapsulate the observation of the error when parsing a class."""
        atok, parse_exception = parse.source_to_atok(source=source)
        assert parse_exception is None, f"{parse_exception=}"
        assert atok is not None

        _, error = parse.atok_to_symbol_table(atok=atok)
        return error


class Test_parse_type_annotation(unittest.TestCase):
    @staticmethod
    def parse_type_annotation_from_ann_assign(
        source: str,
    ) -> Tuple[ast.AST, asttokens.ASTTokens]:
        """Encapsulate the parsing of the type annotation of a variable."""
        atok = asttokens.ASTTokens(source, parse=True)

        module = atok.tree
        assert isinstance(module, ast.Module)
        assert len(module.body) == 1, f"{module.body=}"

        ann_assign = module.body[0]
        assert isinstance(ann_assign, ast.AnnAssign), f"{ann_assign=}"

        assert ann_assign.annotation is not None

        return ann_assign.annotation, atok

    def test_atomic(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: int"
        )

        type_annotation, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is None, tests.common.most_underlying_messages(error)

        self.assertEqual("int", str(type_annotation))

    def test_subscripted_with_a_single_value(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: List[int]"
        )

        type_annotation, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is None, tests.common.most_underlying_messages(error)

        self.assertEqual("List[int]", str(type_annotation))

    def test_subscripted_with_a_tuple(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Mapping[str, Optional[int]]"
        )

        type_annotation, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is None, tests.common.most_underlying_messages(error)

        self.assertEqual("Mapping[str, Optional[int]]", str(type_annotation))

    def test_nested(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Optional[List[Reference]]"
        )

        type_annotation, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is None, tests.common.most_underlying_messages(error)

        self.assertEqual("Optional[List[Reference]]", str(type_annotation))


class Test_parse_type_annotation_fail(unittest.TestCase):
    def test_ellipsis(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Mapping[str, ...]"
        )

        _, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is not None

        self.assertEqual(
            "Expected a string literal if the type annotation is given as a constant, "
            "but got: Ellipsis (as <class 'ellipsis'>)",
            error.message,
        )

    def test_non_name_type_identifier(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: (int if True else str)"
        )

        _, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is not None

        # NOTE (mristin, 2022-01-22):
        # We need to remove the type since it differs between Python 3.8 and newer
        # versions.

        self.assertEqual(
            "Expected either atomic type annotation (as name or string literal) "
            "or a subscripted one (as a subscript), "
            "but got: int if True else str",
            re.sub(r" \(as <class .*>\)$", "", error.message),
        )

    def test_unexpected_slice_in_index(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Optional[str:int]"
        )

        _, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is not None

        self.assertEqual(
            "Expected an index to define a subscripted type annotation, "
            "but got a slice: str:int",
            error.message,
        )

    def test_unexpected_ext_slice_in_index(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Optional[1:2, 3]"
        )

        _, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is not None

        self.assertEqual(
            "Expected an index to define a subscripted type annotation, "
            "but got an extended slice: 1:2, 3",
            error.message,
        )

    def test_unexpected_expression_in_index(self) -> None:
        anno, atok = Test_parse_type_annotation.parse_type_annotation_from_ann_assign(
            "x: Optional[str if True else int]"
        )

        _, error = parse._translate._type_annotation(node=anno, atok=atok)
        assert error is not None

        self.assertEqual(
            "Expected a tuple, a name, a subscript or a string literal "
            "for a subscripted type annotation, but got: str if True else int",
            error.message,
        )


class Test_against_recorded(unittest.TestCase):
    def test_cases(self) -> None:
        repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent
        test_cases_dir = repo_root / "dev/test_data/parse"

        assert test_cases_dir.exists(), f"{test_cases_dir=}"
        assert test_cases_dir.is_dir(), f"{test_cases_dir=}"

        # The expected cases should have no errors.
        expected_pths = sorted((test_cases_dir / "expected").glob("**/meta_model.py"))
        meta_model_pths_expected_errors = [
            (pth, False) for pth in expected_pths
        ]  # type: List[Tuple[pathlib.Path, bool]]

        unexpected_pths = sorted(
            (test_cases_dir / "unexpected").glob("**/meta_model.py")
        )

        meta_model_pths_expected_errors.extend((pth, True) for pth in unexpected_pths)

        for meta_model_pth, expected_errors in meta_model_pths_expected_errors:
            case_dir = meta_model_pth.parent

            try:
                source = meta_model_pth.read_text(encoding="utf-8")

                symbol_table, error = tests.common.parse_source(source)

            except Exception as exception:
                raise AssertionError(
                    f"Expected no exception "
                    f"for the test case {case_dir.relative_to(test_cases_dir)}"
                ) from exception

            if not expected_errors and error is not None:
                raise AssertionError(
                    f"Expected no errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got:\n"
                    f"{tests.common.most_underlying_messages(error)}"
                )

            elif expected_errors and error is None:
                raise AssertionError(
                    f"Expected errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got none."
                )

            else:
                pass

            expected_symbol_table_pth = case_dir / "expected_symbol_table.txt"
            expected_error_pth = case_dir / "expected_error.txt"

            if expected_errors:
                if expected_symbol_table_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded symbol table file when errors "
                        f"are expected: {expected_symbol_table_pth}"
                    )

                assert error is not None

                error_str = tests.common.most_underlying_messages(error)

                if tests.common.RERECORD:
                    expected_error_pth.write_text(error_str, encoding="utf-8")
                else:
                    expected_error_str = expected_error_pth.read_text(encoding="utf-8")
                    self.assertEqual(expected_error_str, error_str, f"{case_dir=}")

            else:
                if expected_error_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded error file when no errors "
                        f"are expected: {expected_error_pth}"
                    )

                assert symbol_table is not None

                symbol_table_str = parse.dump(symbol_table)

                if tests.common.RERECORD:
                    expected_symbol_table_pth.write_text(
                        symbol_table_str, encoding="utf-8"
                    )
                else:
                    expected_symbol_table_str = expected_symbol_table_pth.read_text(
                        encoding="utf-8"
                    )
                    self.assertEqual(
                        expected_symbol_table_str,
                        symbol_table_str,
                        f"{case_dir=}, {error=}",
                    )

    def test_real_meta_models(self) -> None:
        repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent
        test_cases_dir = repo_root / "dev/test_data/parse/real_meta_models"

        assert test_cases_dir.exists(), f"{test_cases_dir=}"
        assert test_cases_dir.is_dir(), f"{test_cases_dir=}"

        for meta_model_pth in tests.common.REAL_META_MODEL_PATHS:
            case_dir = test_cases_dir / meta_model_pth.stem

            try:
                source = meta_model_pth.read_text(encoding="utf-8")

                symbol_table, error = tests.common.parse_source(source)

            except Exception as exception:
                raise AssertionError(
                    f"Expected no exception when parsing "
                    f"the real meta-model {meta_model_pth}"
                ) from exception

            if error is not None:
                raise AssertionError(
                    f"Expected no errors when parsing "
                    f"the real meta-model {meta_model_pth}, but got:\n"
                    f"{tests.common.most_underlying_messages(error)}"
                )

            expected_symbol_table_pth = case_dir / "expected_symbol_table.txt"

            assert symbol_table is not None

            symbol_table_str = parse.dump(symbol_table)

            if tests.common.RERECORD:
                expected_symbol_table_pth.write_text(symbol_table_str, encoding="utf-8")
            else:
                expected_symbol_table_str = expected_symbol_table_pth.read_text(
                    encoding="utf-8"
                )
                self.assertEqual(
                    expected_symbol_table_str,
                    symbol_table_str,
                    f"{case_dir=}, {error=}",
                )


if __name__ == "__main__":
    unittest.main()



"""Compare the rendered regexes against their original form in ``source.py``."""

import os
import pathlib
import sys
from typing import Optional


def main() -> int:
    """Execute the main routine."""

    repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent
    test_cases_dir = repo_dir / "dev/test_data/parse_retree"
    expected_pths = sorted((test_cases_dir / "expected").glob("**/source.py"))

    for pth in expected_pths:
        rendered_regex_pth = pth.parent / "rendered_regex.txt"

        source = pth.read_text(encoding="utf-8").strip()

        source_stem = None  # type: Optional[str]
        if source.startswith('"') and source.endswith('"'):
            source_stem = source[1:-1]
        elif source.startswith('f"') and source.endswith('"'):
            source_stem = source[2:-1]
        else:
            pass

        rendered = rendered_regex_pth.read_text(encoding="utf-8")
        rendered_repr_stem = repr(rendered)[1:-1]

        if source_stem is None or rendered_repr_stem != source_stem:
            print()
            print(str(pth))
            print(source)
            print(repr(rendered))

    return 0


if __name__ == "__main__":
    sys.exit(main())



class SomethingAbstract:
    property_a: str
    property_b: Optional[str]

    def __init__(
        self,
        property_a: str,
        # The default to None is missing here.
        property_b: Optional[str] = "abc",
    ) -> None:
        self.property_a = property_a
        self.property_b = property_b


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import unittest

from aas_core_codegen.yielding import (
    flow as yielding_flow,
    linear as yielding_linear,
)


class TestLinearization(unittest.TestCase):
    def test_empty(self) -> None:
        statements = yielding_linear._linearize_control_flow(flow=[])

        dump = yielding_linear.dump(statements)
        self.assertEqual("", dump)

    def test_command(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[yielding_flow.command_from_text("DoSomething()")]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: DoSomething()""",
            dump,
        )

    def test_yield(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[yielding_flow.Yield()]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual("0: yield", dump)

    def test_simple_if_true_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                    or_else=[yielding_flow.command_from_text("DoSomethingElse()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if CheckSomething()
   is false, jump to 3
1: DoSomething()
2: jump 4
3: DoSomethingElse()
4: noop
5: Finalize()""",
            dump,
        )

    def test_simple_if_true_without_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if CheckSomething()
   is false, jump to 2
1: DoSomething()
2: noop
3: Finalize()""",
            dump,
        )

    def test_nested_if_true(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckOuter()",
                    body=[
                        yielding_flow.IfTrue(
                            condition="OuterTrue_CheckInner()",
                            body=[
                                yielding_flow.command_from_text("OuterTrue_DoInner()")
                            ],
                            or_else=[
                                yielding_flow.command_from_text(
                                    "OuterTrue_DoElseInner()"
                                )
                            ],
                        )
                    ],
                    or_else=[
                        yielding_flow.IfTrue(
                            condition="OuterFalse_CheckInner()",
                            body=[
                                yielding_flow.command_from_text("OuterFalse_DoInner()")
                            ],
                            or_else=[
                                yielding_flow.command_from_text(
                                    "OuterFalse_DoElseInner()"
                                )
                            ],
                        )
                    ],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
 0: if CheckOuter()
    is false, jump to 7
 1: if OuterTrue_CheckInner()
    is false, jump to 4
 2: OuterTrue_DoInner()
 3: jump 5
 4: OuterTrue_DoElseInner()
 5: noop
 6: jump 12
 7: if OuterFalse_CheckInner()
    is false, jump to 10
 8: OuterFalse_DoInner()
 9: jump 11
10: OuterFalse_DoElseInner()
11: noop
12: noop
13: Finalize()""",
            dump,
        )

    def test_simple_if_false_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfFalse(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSinceCheckFalse()")],
                    or_else=[yielding_flow.command_from_text("DoSinceCheckTrue()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if CheckSomething()
   is true, jump to 3
1: DoSinceCheckFalse()
2: jump 4
3: DoSinceCheckTrue()
4: noop
5: Finalize()""",
            dump,
        )

    def test_simple_if_false_without_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfFalse(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSinceCheckFalse()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if CheckSomething()
   is true, jump to 2
1: DoSinceCheckFalse()
2: noop
3: Finalize()""",
            dump,
        )

    def test_for(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.For(
                    init="Start()",
                    condition="!Done()",
                    iteration="Next()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: Start()
1: if !Done()
   is false, jump to 5
2: DoSomething()
3: Next()
4: jump 1
5: noop
6: Finalize()""",
            dump,
        )

    def test_for_without_init(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.For(
                    condition="!Done()",
                    iteration="Next()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if !Done()
   is false, jump to 4
1: DoSomething()
2: Next()
3: jump 0
4: noop
5: Finalize()""",
            dump,
        )

    def test_nested_if_in_for(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.For(
                    init="Start()",
                    condition="!Done()",
                    iteration="Next()",
                    body=[
                        yielding_flow.IfTrue(
                            condition="Check()",
                            body=[yielding_flow.command_from_text("DoSomething()")],
                        )
                    ],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: Start()
1: if !Done()
   is false, jump to 7
2: if Check()
   is false, jump to 4
3: DoSomething()
4: noop
5: Next()
6: jump 1
7: noop
8: Finalize()""",
            dump,
        )

    def test_while(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.While(
                    condition="!Done()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        dump = yielding_linear.dump(statements)
        self.assertEqual(
            """\
0: if !Done()
   is false, jump to 3
1: DoSomething()
2: jump 0
3: noop
4: Finalize()""",
            dump,
        )


class TestCompression(unittest.TestCase):
    def test_simple_if_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                    or_else=[yielding_flow.command_from_text("DoSomethingElse()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        statements = yielding_linear._compress_in_place(statements)
        dump = yielding_linear.dump(statements)

        self.assertEqual(
            """\
 : if CheckSomething()
   is false, jump to 3
 : DoSomething()
 : jump 4
3: DoSomethingElse()
4: Finalize()""",
            dump,
        )

    def test_simple_if_without_else(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckSomething()",
                    body=[yielding_flow.command_from_text("DoSomething()")],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        statements = yielding_linear._compress_in_place(statements)
        dump = yielding_linear.dump(statements)

        self.assertEqual(
            """\
 : if CheckSomething()
   is false, jump to 2
 : DoSomething()
2: Finalize()""",
            dump,
        )

    def test_nested_if(self) -> None:
        statements = yielding_linear._linearize_control_flow(
            flow=[
                yielding_flow.IfTrue(
                    condition="CheckOuter()",
                    body=[
                        yielding_flow.IfTrue(
                            condition="OuterTrue_CheckInner()",
                            body=[
                                yielding_flow.command_from_text("OuterTrue_DoInner()")
                            ],
                            or_else=[
                                yielding_flow.command_from_text(
                                    "OuterTrue_DoElseInner()"
                                )
                            ],
                        )
                    ],
                    or_else=[
                        yielding_flow.IfTrue(
                            condition="OuterFalse_CheckInner()",
                            body=[
                                yielding_flow.command_from_text("OuterFalse_DoInner()")
                            ],
                            or_else=[
                                yielding_flow.command_from_text(
                                    "OuterFalse_DoElseInner()"
                                )
                            ],
                        )
                    ],
                ),
                yielding_flow.command_from_text("Finalize()"),
            ]
        )

        statements = yielding_linear._compress_in_place(statements)
        dump = yielding_linear.dump(statements)

        self.assertEqual(
            """\
  : if CheckOuter()
    is false, jump to 7
  : if OuterTrue_CheckInner()
    is false, jump to 4
  : OuterTrue_DoInner()
  : jump 5
 4: OuterTrue_DoElseInner()
 5: jump 11
 7: if OuterFalse_CheckInner()
    is false, jump to 10
  : OuterFalse_DoInner()
  : jump 11
10: OuterFalse_DoElseInner()
11: Finalize()""",
            dump,
        )


class TestLinearizeToSubroutines(unittest.TestCase):
    def test_empty(self) -> None:
        result = yielding_linear.linearize_to_subroutines(flow=[])
        self.assertListEqual([], result)

    def test_inspired_by_verificator_in_cpp(self) -> None:
        # NOTE (mristin, 2023-10-21):
        # We test inspired by a real-world example from C++ code generation so that
        # we can see if it works, estimate how ergonomic it is to write the flows, and
        # check whether the resulting code is readable.
        #
        # Hence, we test with two invariants and two constrained primitive properties
        # so that all the above can be tested.
        subroutines = yielding_linear.linearize_to_subroutines(
            flow=[
                yielding_flow.IfTrue(
                    condition="Invariant0Invalidated",
                    body=[
                        yielding_flow.command_from_text(
                            "error_ = ErrorForInvariant0()"
                        ),
                        yielding_flow.Yield(),
                    ],
                ),
                yielding_flow.IfTrue(
                    condition="Invariant1Invalidated",
                    body=[
                        yielding_flow.command_from_text(
                            "error_ = ErrorForInvariant1()"
                        ),
                        yielding_flow.Yield(),
                    ],
                ),
                yielding_flow.IfTrue(
                    condition="instance_.prop0.HasValue()",
                    body=[
                        yielding_flow.command_from_text(
                            """\
verification = VerifyConstrainedPrimitive(instance_.prop0.Value())
iterator_ = verification.Begin()
iterator_end_ = verification.End()"""
                        ),
                        yielding_flow.For(
                            condition="iterator_ != iterator_end_",
                            iteration="++iterator_",
                            body=[
                                yielding_flow.command_from_text("error_ = *iterator_"),
                                yielding_flow.Yield(),
                            ],
                        ),
                        yielding_flow.command_from_text(
                            """\
iterator_.reset()
iterator_end_.reset()"""
                        ),
                    ],
                ),
                yielding_flow.IfTrue(
                    condition="instance_.prop1.HasValue()",
                    body=[
                        yielding_flow.command_from_text(
                            """\
verification = VerifyConstrainedPrimitive(instance_.prop1.Value())
iterator_ = verification.Begin()
iterator_end_ = verification.End()"""
                        ),
                        yielding_flow.For(
                            condition="iterator_ != iterator_end_",
                            iteration="++iterator_",
                            body=[
                                yielding_flow.command_from_text("error_ = *iterator_"),
                                yielding_flow.Yield(),
                            ],
                        ),
                        yielding_flow.command_from_text(
                            """\
iterator_.reset()
iterator_end_.reset()"""
                        ),
                    ],
                ),
                yielding_flow.command_from_text(
                    """\
error_.Reset()
Finalize()"""
                ),
            ]
        )

        output = "\n---\n".join(
            yielding_linear.dump(subroutine) for subroutine in subroutines
        )

        self.maxDiff = None
        self.assertEqual(
            """\
0: if Invariant0Invalidated
   is false, jump to 1
 : error_ = ErrorForInvariant0()
 : yield
---
1: if Invariant1Invalidated
   is false, jump to 2
 : error_ = ErrorForInvariant1()
 : yield
---
2: if instance_.prop0.HasValue()
   is false, jump to 6
 : verification = VerifyConstrainedPrimitive(instance_.prop0.Value())
   iterator_ = verification.Begin()
   iterator_end_ = verification.End()
---
3: if iterator_ != iterator_end_
   is false, jump to 5
 : error_ = *iterator_
 : yield
---
4: ++iterator_
 : jump 3
---
5: iterator_.reset()
   iterator_end_.reset()
---
6: if instance_.prop1.HasValue()
   is false, jump to 10
 : verification = VerifyConstrainedPrimitive(instance_.prop1.Value())
   iterator_ = verification.Begin()
   iterator_end_ = verification.End()
---
7: if iterator_ != iterator_end_
   is false, jump to 9
 : error_ = *iterator_
 : yield
---
8: ++iterator_
 : jump 7
---
9: iterator_.reset()
   iterator_end_.reset()
---
10: error_.Reset()
    Finalize()""",
            output,
        )


if __name__ == "__main__":
    unittest.main()



"""
Do something.

    0: An unexpected block quote.
"""


class Something:
    """
    Represent something.

        1: An unexpected block quote.
    """

    something: str
    """
    Capture something.
    
        2: An unexpected block quote.
    """

    def __init__(self, something: str) -> None:
        self.something = something

    def do_something(self) -> None:
        """
        Do something.

            3: An unexpected block quote.
        """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Something:
    property_a: str
    property_b: str

    property_c: Optional[str]
    property_d: Optional[str]

    def __init__(
        self,
        property_b: str,
        property_a: str,
        property_d: Optional[str] = None,
        property_c: Optional[str] = None,
    ) -> None:
        self.property_a = property_a
        self.property_b = property_b
        self.property_c = property_c
        self.property_d = property_d


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# fmt: off
@invariant(
    lambda self:
    some_unexpected_builtin_function(self.some_property),
    "Some property must abide by some unexpected builtin function."
)
# fmt: on
class Something:
    some_property: str

    def __init__(self, some_property: str) -> None:
        self.some_property = some_property


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Something:
    some_property: Optional[str]

    def __init__(self, some_property: Optional[str] = None) -> None:
        self.some_property = some_property


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import unittest
from typing import List

from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.cpp.verification import _generate as cpp_verification_generate
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from tests import common as tests_common


class Test_against_recorded(unittest.TestCase):
    def test_optional_vector_in_invariant_not_deferenced(self) -> None:
        # NOTE (mristin):
        # This is a regression test where we check that an optional list without
        # implication is not de-referenced in the invariant's transpiled code.

        source = """\
class Item:
    value: str
    
    def __init__(self, value: str) -> None:
        self.value = value

@verification
@implementation_specific
def check_something(value: Optional[List[Item]]) -> bool:
    pass

@invariant(
    lambda self:
    check_something(self.value),
    "Some description"
)
class Something:
    value: Optional[List[Item]]
    
    def __init__(self, value: Optional[List[Item]] = None) -> None:
        self.value = value
        
__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""

        symbol_table = tests_common.must_translate_source_to_intermediate(source=source)

        base_environment = intermediate_type_inference.populate_base_environment(
            symbol_table=symbol_table
        )

        something_cls = symbol_table.must_find_concrete_class(
            name=Identifier("Something")
        )

        environment = intermediate_type_inference.MutableEnvironment(
            parent=base_environment
        )
        environment.set(
            identifier=Identifier("self"),
            type_annotation=intermediate_type_inference.OurTypeAnnotation(
                our_type=something_cls
            ),
        )

        blocks = []  # type: List[Stripped]
        for invariant in something_cls.invariants:
            (
                condition_expr,
                error,
            ) = cpp_verification_generate._transpile_class_invariant(
                invariant=invariant, symbol_table=symbol_table, environment=environment
            )
            assert error is None, (
                f"Unexpected generation error for an invariant: "
                f"{tests_common.most_underlying_messages(error)}"
            )
            assert condition_expr is not None

            blocks.append(condition_expr)

        assert len(blocks) == 1, (
            f"Expected only a single block for a single invariant "
            f"in the class {something_cls.name!r}"
        )

        # NOTE (mristin):
        # The implementation of ``CheckSomething`` needs to deal with the optional
        # values. As soon as something is optional in C++, it will be provided to
        # the function as-is. This is intentional.
        self.assertEqual(
            """\
CheckSomething(
  instance_->value()
)""",
            blocks[0],
        )

    def test_optional_in_invariant_dereferenced_if_certainly_not_null(
        self,
    ) -> None:
        # NOTE (mristin):
        # This is a regression test where we check that an optional value is correctly
        # de-referenced in the invariant's transpiled code if type inference determines
        # that it is certainly not null.

        source = """\
class Item:
    value: str

    def __init__(self, value: str) -> None:
        self.value = value

@verification
@implementation_specific
def check_something(value: Optional[List[Item]]) -> bool:
    pass

@invariant(
    lambda self:
    (self.value is None) or check_something(self.value),
    "Some description"
)
class Something:
    value: Optional[List[Item]]

    def __init__(self, value: Optional[List[Item]] = None) -> None:
        self.value = value

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""

        symbol_table = tests_common.must_translate_source_to_intermediate(source=source)

        base_environment = intermediate_type_inference.populate_base_environment(
            symbol_table=symbol_table
        )

        something_cls = symbol_table.must_find_concrete_class(
            name=Identifier("Something")
        )

        environment = intermediate_type_inference.MutableEnvironment(
            parent=base_environment
        )
        environment.set(
            identifier=Identifier("self"),
            type_annotation=intermediate_type_inference.OurTypeAnnotation(
                our_type=something_cls
            ),
        )

        blocks = []  # type: List[Stripped]
        for invariant in something_cls.invariants:
            (
                condition_expr,
                error,
            ) = cpp_verification_generate._transpile_class_invariant(
                invariant=invariant, symbol_table=symbol_table, environment=environment
            )
            assert error is None, (
                f"Unexpected generation error for an invariant: "
                f"{tests_common.most_underlying_messages(error)}"
            )
            assert condition_expr is not None

            blocks.append(condition_expr)

        assert len(blocks) == 1, (
            f"Expected only a single block for a single invariant "
            f"in the class {something_cls.name!r}"
        )

        # NOTE (mristin):
        # If we know that the value is not optional, we explicitly de-reference it.
        self.assertEqual(
            """\
(
  (!(instance_->value().has_value()))
  || CheckSomething(
    (*(instance_->value()))
  )
)""",
            blocks[0],
        )


if __name__ == "__main__":
    unittest.main()



class Some_enum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"
    Yet_another_literal = "YET-ANOTHER-LITERAL"


Some_set: Set[Some_enum] = constant_set(
    values=[
        Some_enum.Some_literal,
        Some_enum.Another_literal,
        Some_enum.Yet_another_literal,
    ]
)

Another_set: Set[Some_enum] = constant_set(
    values=[Some_enum.Some_literal, Some_enum.Another_literal], superset_of=[Some_set]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
